"""This file contains paths to the parameter files.\
    Please refer to the readme on where to get the parameters.\
        Save them in this folder: ~/.zerodose_params."""
import os


# please refer to the readme on where to get the parameters. Save them in this folder:
folder_with_parameter_files = os.path.join(
    os.path.expanduser("~"), ".zerodose_data", "model_params"
)
folder_with_templates = os.path.join(
    os.path.expanduser("~"), ".zerodose_data", "templates"
)



"""Pipeline for ZeroDose project."""
from .abnormality import create_abnormality_maps
from .full import run_full
from .niftyreg import from_mni
from .niftyreg import to_mni
from .normalization import normalize_to_pet
from .synthetization import synthesize_baselines


__all__ = [
    "create_abnormality_maps",
    "run_full",
    "from_mni",
    "to_mni",
    "normalize_to_pet",
    "synthesize_baselines",
]



"""Dataset class for the ZeroDose project."""
from typing import Any
from typing import Dict

import torchio as tio

from zerodose.processing import Binarize
from zerodose.processing import PadAndCropMNI
from zerodose.processing import ToFloat32


class SubjectDataset(tio.data.SubjectsDataset):
    """Dataset class for the ZeroDose project."""

    def __init__(self, mri_fnames, mask_fnames, out_fnames):
        """Initialize the dataset."""
        transforms = self._get_augmentation_transform_val()

        subjects = [
            self._make_subject_predict(mr_f, ma_f, ou_f)
            for mr_f, ma_f, ou_f in zip(  # noqa
                mri_fnames,
                mask_fnames,
                out_fnames,
            )
        ]
        super().__init__(subjects, transforms)

    def _make_subject_dict(self, mr_path, mask_path) -> dict:
        subject_dict: Dict[Any, Any] = {}
        mri = mr_path
        mask = mask_path

        subject_dict["mr"] = tio.ScalarImage(mri)
        subject_dict["mask"] = tio.LabelMap(mask)

        return subject_dict

    def _make_subject_predict(self, mr_path, mask_path, out_fname) -> tio.Subject:
        subject_dict = self._make_subject_dict(mr_path, mask_path)
        subject_dict["out_fname"] = out_fname
        return tio.Subject(subject_dict)

    def _get_augmentation_transform_val(self) -> tio.Compose:
        augmentations = [
            Binarize(include=["mask"]),
            tio.transforms.ZNormalization(include=["mr"], masking_method="mask"),
            PadAndCropMNI(),
            ToFloat32(include=["mr"]),
        ]

        return tio.Compose(augmentations)



"""Networks for ZeroDose."""
from typing import List

import torch
import torch.nn as nn


class ConvBlock(nn.Module):
    """Convolutional block."""

    def __init__(self, ch_i: int, ch_o: int, dropout: float = 0) -> None:
        """Initialize the block."""
        self.sequence: List[nn.Module] = []
        super().__init__()

        for _i in range(2):
            self.sequence += [nn.Conv3d(ch_i, ch_o, padding=1, kernel_size=3)]
            self.sequence += [nn.BatchNorm3d(ch_o)]
            self.sequence += [nn.ReLU(True)]
            if dropout:
                self.sequence += [nn.Dropout(dropout)]

            ch_i = ch_o
        self.model = nn.Sequential(*self.sequence)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the block."""
        return self.model(x)


class UpConv(nn.Module):
    """Upsampling convolutional block."""

    def __init__(self, ch_i: int, ch_o: int) -> None:
        """Initialize the block."""
        self.sequence: List[nn.Module] = []
        super().__init__()

        self.sequence += [
            nn.ConvTranspose3d(
                ch_i, ch_o, stride=2, padding=1, kernel_size=3, output_padding=1
            )
        ]
        self.sequence += [nn.BatchNorm3d(ch_o)]
        self.sequence += [nn.ReLU(True)]

        self.model = nn.Sequential(*self.sequence)

    def forward(self, x: torch.Tensor, skip: torch.Tensor) -> torch.Tensor:
        """Forward pass of the block."""
        a: torch.Tensor = self.model(x)

        return torch.cat((skip, a), dim=1)


class DownConv(nn.Module):
    """Downsampling convolutional block."""

    def __init__(self, ch_i: int, ch_o: int) -> None:
        """Initialize the block."""
        self.sequence: List[nn.Module] = []
        super().__init__()
        self.sequence += [nn.Conv3d(ch_i, ch_o, stride=2, padding=1, kernel_size=3)]
        self.sequence += [nn.BatchNorm3d(ch_o)]
        self.sequence += [nn.LeakyReLU(0.2, True)]
        self.model = nn.Sequential(*self.sequence)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the block."""
        return self.model(x)


class UNet3D(nn.Module):
    """3D U-Net model for PET reconstruction."""

    def __init__(self, use_dropout: bool = False) -> None:
        """Initialize the model."""
        super().__init__()

        self.conv_d1 = ConvBlock(1, 32, dropout=0.1 * use_dropout)  # 32, 196, 196

        self.down_1 = DownConv(32, 128)  # 16, 48, 48

        self.conv_d2 = ConvBlock(128, 64, dropout=0.1 * use_dropout)  # 16, 48, 48
        self.down_2 = DownConv(64, 256)  # 8, 24, 24

        self.conv_d3 = ConvBlock(256, 128, dropout=0.2 * use_dropout)  # 8, 24, 24
        self.down_3 = DownConv(128, 512)  # 4, 12, 12

        self.conv_d4 = ConvBlock(512, 256, dropout=0.2 * use_dropout)  # 4, 12, 12
        self.down_4 = DownConv(256, 512)  # 2, 6, 6

        # Bottom
        self.conv_b = ConvBlock(512, 512, dropout=0.3 * use_dropout)  # 2, 6, 6

        # Going up
        self.up_1 = UpConv(512, 256)  # 4, 12, 12
        self.conv_u1 = ConvBlock(512, 256, dropout=0.3 * use_dropout)  # 4, 12, 12

        self.up_2 = UpConv(256, 128)  # 8, 24, 24
        self.conv_u2 = ConvBlock(256, 128, dropout=0.2 * use_dropout)  # 8, 24, 24

        self.up_3 = UpConv(128, 64)  # 16, 48, 48
        self.conv_u3 = ConvBlock(128, 64, dropout=0.2 * use_dropout)  # 16, 48, 48

        self.up_4 = UpConv(64, 32)  # 32, 196, 196
        self.conv_u4 = ConvBlock(64, 32, dropout=0.1 * use_dropout)  # 32, 196, 196

        self.last = nn.Conv3d(32, 1, padding=1, kernel_size=3)  # 32, 196, 196
        self.last_sigmoid = nn.Sigmoid()  # 32, 196, 196

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the model."""
        # Going down
        c1 = self.conv_d1(x)
        x = self.down_1(c1)

        c2 = self.conv_d2(x)
        x = self.down_2(c2)

        c3 = self.conv_d3(x)
        x = self.down_3(c3)

        c4 = self.conv_d4(x)
        x = self.down_4(c4)

        # Bottom
        x = self.conv_b(x)

        # Going up
        x = self.conv_u1(self.up_1(x, c4))
        x = self.conv_u2(self.up_2(x, c3))
        x = self.conv_u3(self.up_3(x, c2))
        x = self.conv_u4(self.up_4(x, c1))

        x = self.last(x)
        return x


class DummyGenerator(nn.Module):
    """Dummy generator for testing purposes."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the generator."""
        super().__init__()
        self.w = nn.Parameter(torch.tensor(1.0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the generator."""
        return self.w * x



"""Model definition for ZeroDose."""
import torch
import torch.nn as nn

from zerodose import utils

from .networks import DummyGenerator
from .networks import UNet3D


class ZeroDose(nn.Module):
    """ZeroDose model."""

    def __init__(self, model_type: str = "standard") -> None:
        """Initialize the model."""
        self.model_type = model_type
        self.generator: nn.Module
        super().__init__()
        if model_type == "standard":
            self.generator = UNet3D(use_dropout=True)
        elif model_type == "dummy":
            self.generator = DummyGenerator()
        else:
            raise ValueError(f"Model type {model_type} not recognized.")

    def forward(self, mrs: torch.Tensor) -> torch.Tensor:
        """Forward pass through the model."""
        return self.generator(mrs)


class AbnormalityMap(nn.Module):
    """Abnormality map generator."""

    def __init__(self, sigma_smooth=3) -> None:
        """Initialize the Abnormality map."""
        super().__init__()
        self.smooth = utils.GaussianSmoothing(
            channels=1, kernel_size=5 * sigma_smooth, sigma=sigma_smooth, dim=3
        )

    def forward(
        self, pet: torch.Tensor, sbpet: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Generate the abnormality map."""
        pet_blurred = self.smooth(pet)
        sbpet_blurred = self.smooth(sbpet)
        abnormality_map = (pet_blurred - sbpet_blurred) / (sbpet_blurred + 1e-7) * 100
        abnormality_map[torch.isnan(abnormality_map)] = 0
        abnormality_map *= mask
        return abnormality_map


class QuantileNormalization(nn.Module):
    """Quantile normalization of the sbPET image."""

    def __init__(
        self,
        quantile,
        sigma_normalization=3,
    ) -> None:
        """Initialize the normalization module."""
        super().__init__()
        self.quantile = quantile
        self.smooth_normalization = utils.GaussianSmoothing(
            channels=1,
            kernel_size=5 * sigma_normalization,
            sigma=sigma_normalization,
            dim=3,
        )

    def forward(
        self, pet: torch.Tensor, sbpet: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Normalize the sbPET image."""
        return self._scale_sbpet(pet, sbpet, mask)

    def _get_normalization_mask(self, pet, mask):
        qt = torch.quantile(pet[mask], self.quantile)
        norm_mask = mask & (pet > qt)
        return norm_mask

    def _scale_sbpet(self, pet, sbpet, mask):
        for _i in range(2):
            pet_blurred = self.smooth_normalization(pet)
            norm_mask = self._get_normalization_mask(pet_blurred, mask)
            norm_const = torch.mean(pet[norm_mask]) / torch.mean(sbpet[norm_mask])
            sbpet *= norm_const
            sbpet[~mask] = pet[~mask]

        return sbpet



"""Full pipeline."""

import os
import shutil
import tempfile

from zerodose.pipeline.abnormality import create_abnormality_maps
from zerodose.pipeline.niftyreg import from_mni
from zerodose.pipeline.niftyreg import to_mni
from zerodose.pipeline.normalization import normalize_to_pet
from zerodose.pipeline.synthetization import synthesize_baselines


def run_full(  # noqa: C901
    mri_fname,
    mask_fname,
    out_sbpet,
    pet_fname=None,
    out_abn=None,
    out_img=None,
    do_registration=True,
    do_abnormality=True,
    do_normalization=True,
    do_image=True,
    verbose=False,
    device="cuda:0",
    outputspace="pet",
):
    """Run full pipeline."""
    with tempfile.TemporaryDirectory() as tmpdir:
        if pet_fname is None:
            print("No pet file provided:")
            print("\tsetting outputspace = 'mr'")
            print("\tskipping normalization")
            print("\tskipping abnormality map")
            print("\tSkipping image creation")
            outputspace = "mr"
            do_normalization = False
            do_abnormality = False
            do_image = False

        if outputspace == "pet":
            output_target = pet_fname
        elif outputspace == "mr":
            output_target = mri_fname

        sbpetraw = os.path.join(tmpdir, "sbpetraw.nii.gz")
        sbpet = os.path.join(tmpdir, "sbpet.nii.gz")

        if do_registration:
            affine_fname = os.path.join(tmpdir, "affine.txt")
            mri = os.path.join(tmpdir, "mri_mni.nii.gz")
            pet = os.path.join(tmpdir, "pet_mni.nii.gz")
            mask = os.path.join(tmpdir, "mask_mni.nii.gz")
            if verbose:
                print("Registration to MNI space...")
            to_mni(
                in_mri_fname=mri_fname,
                in_pet_fname=pet_fname,
                in_mask_fname=mask_fname,
                out_mri_fname=mri,
                out_pet_fname=pet,
                out_mask_fname=mask,
                out_affine_fname=affine_fname,
            )
        else:
            mri = mri_fname
            pet = pet_fname
            mask = mask_fname

        synthesize_baselines(
            mri_fnames=mri,
            mask_fnames=mask,
            out_fnames=sbpetraw,
            verbose=verbose,
            device=device,
        )

        if do_normalization:
            normalize_to_pet(
                pet_fnames=pet,
                sbpet_fnames=sbpetraw,
                mask_fnames=mask,
                out_fnames=sbpet,
                verbose=verbose,
                device=device,
            )
        else:
            sbpet = sbpetraw

        if do_registration:
            if verbose:
                print(f"Registration of sbPET back to {outputspace} space...")
            from_mni(
                ref_fname=output_target,
                in_affine_fname=affine_fname,
                float_fname=sbpet,
                out_fname=out_sbpet,
            )
        else:
            shutil.copyfile(sbpet, out_sbpet)

        if verbose:
            print("sbPET saved to", out_sbpet)

        if do_abnormality:
            abn = os.path.join(tmpdir, "abn.nii.gz")

            create_abnormality_maps(
                pet_fnames=pet,
                sbpet_fnames=sbpet,
                mask_fnames=mask,
                out_fnames=abn,
                verbose=verbose,
                device=device,
            )

            if do_image:
                create_abn_image(
                    mri=mri,
                    pet=pet,
                    sbpet=sbpet,
                    abn=abn,
                    mask=mask,
                    out_fname=out_img,
                )

            if do_registration:
                if verbose:
                    print(
                        f"Registration of abnormality map back\
                              to {outputspace} space..."
                    )
                from_mni(
                    ref_fname=output_target,
                    in_affine_fname=affine_fname,
                    float_fname=abn,
                    out_fname=out_abn,
                )
            else:
                shutil.copyfile(abn, out_abn)
            if verbose:
                print("Abnormality map saved to", out_abn)


def create_abn_image(mri, pet, sbpet, abn, mask, out_fname):
    """Create abnormality image."""
    pass



"""Functions called by the CLI."""


from typing import Sequence

import torch

from zerodose import utils
from zerodose.models import AbnormalityMap


def create_abnormality_maps(
    pet_fnames: Sequence[str],
    sbpet_fnames: Sequence[str],
    mask_fnames: Sequence[str],
    out_fnames: Sequence[str],
    verbose: bool = False,
    save_output: bool = True,
    device: str = "cuda:0",
) -> None:
    """Create abnormality maps from PET and sbPET images."""
    if isinstance(pet_fnames, str):
        pet_fnames = [pet_fnames]
    if isinstance(sbpet_fnames, str):
        sbpet_fnames = [sbpet_fnames]
    if isinstance(mask_fnames, str):
        mask_fnames = [mask_fnames]
    if isinstance(out_fnames, str):
        out_fnames = [out_fnames]

    if not (
        len(pet_fnames) == len(mask_fnames)
        and len(pet_fnames) == len(sbpet_fnames)
        and len(pet_fnames) == len(out_fnames)
    ):
        raise ValueError(
            """The number of PET files {} sbPET files {},
            mask files {} and output files {} must be identical""".format(
                len(pet_fnames), len(sbpet_fnames), len(mask_fnames), len(out_fnames)
            )
        )

    abnormality_mapper = AbnormalityMap(sigma_smooth=3).to(device)

    with torch.no_grad():
        for i in range(len(pet_fnames)):  # type: ignore
            if verbose:
                print(f"Creating abnormality map for {pet_fnames[i]}")

            pet = utils.load_nifty(pet_fnames[i]).to(device)
            sbpet = utils.load_nifty(sbpet_fnames[i]).to(device)
            mask = utils.load_nifty(mask_fnames[i]).type(torch.bool).to(device)

            abnormality_map = abnormality_mapper(pet, sbpet, mask)

            if save_output:
                if verbose:
                    print(f"Saving to {out_fnames[i]}")

                utils.save_nifty(
                    abnormality_map, out_fnames[i], affine_ref=str(pet_fnames[i])
                )



"""Functions called by the CLI."""


from typing import Sequence
from typing import Union

import torch
import torchio as tio

from zerodose import utils
from zerodose.dataset import SubjectDataset
from zerodose.inference import _infer_single_subject


def synthesize_baselines(
    mri_fnames: Sequence[str],
    mask_fnames: Sequence[str],
    out_fnames: Sequence[str],
    device: Union[torch.device, str] = "cuda:0",
    sd_weight: Union[float, int] = 5,
    verbose: bool = False,
    batch_size: int = 1,
    stride: int = 2,
    save_output: bool = True,
    apply_mask: bool = True,
) -> None:
    """Synthesize baseline PET images from MRI images."""
    if isinstance(mri_fnames, str):
        mri_fnames = [mri_fnames]
    if isinstance(mask_fnames, str):
        mask_fnames = [mask_fnames]
    if isinstance(out_fnames, str):
        out_fnames = [out_fnames]

    if not (len(mri_fnames) == len(mask_fnames) and len(mri_fnames) == len(out_fnames)):
        raise Exception(
            """The number of input files {} mask files {}
            and output files {} must be identical""".format(
                len(mri_fnames), len(mask_fnames), len(out_fnames)
            )
        )

    dataset = SubjectDataset(mri_fnames, mask_fnames, out_fnames)
    model = utils.get_model()

    model = model.to(device)
    model.eval()
    patch_size = (32, 192, 192)
    patch_overlap = (32 - stride, 192 - stride, 192 - stride)

    for sub in dataset:  # type: ignore
        if verbose:
            print(f"Synthesizing sbPET for {sub.mr.path}")

        sbpet_tensor = _infer_single_subject(
            model, sub, patch_size, patch_overlap, batch_size, sd_weight, device
        )

        sbpet_img = tio.ScalarImage(
            tensor=sbpet_tensor.cpu().data, affine=sub["mr"].affine
        )

        inverse_transform = sub.get_inverse_transform(warn=False)

        sub.add_image(sbpet_img, "sbpet")
        sub.remove_image("mr")
        sub = inverse_transform(sub)

        if save_output:
            if verbose:
                print(f"Saving to {sub.out_fname}")
            sub["sbpet"].save(sub.out_fname)



"""Command-line interface."""
from typing import Sequence
from typing import Union

import click

from zerodose.pipeline import create_abnormality_maps
from zerodose.pipeline import normalize_to_pet
from zerodose.pipeline import run_full
from zerodose.pipeline import synthesize_baselines
from zerodose.utils import _create_output_fname


@click.group()
@click.version_option()
def main() -> None:
    """Zerodose CLI."""
    pass


device_option = click.option(
    "-d",
    "--device",
    "device",
    type=click.Choice(
        [
            "cpu",
            "cuda:0",
            "cuda:1",
            "cuda:2",
            "cuda:3",
            "cuda:4",
            "cuda:5",
            "cuda:6",
            "cuda:7",
        ]
    ),
    default="cuda:0",
    help="Device to use for inference.",
)

mri_option = click.option(
    "-i",
    "--in",
    "mri_fnames",
    type=click.Path(exists=True),
    required=True,
    multiple=True,
    help="Help test",
)

mask_option = click.option(
    "-m",
    "--mask",
    "mask_fnames",
    type=click.Path(exists=True),
    required=True,
    multiple=True,
)

sbpet_output_option = click.option(
    "-o", "--out", "out_fnames", type=click.Path(), multiple=True
)

verbose_option = click.option(
    "-v",
    "--verbose",
    "verbose",
    is_flag=True,
    default=False,
    help="Print verbose output.",
)

no_registration_option = click.option(
    "-n",
    "--no-registration",
    "no_registration",
    is_flag=True,
    default=False,
    help="""Skip registration to MNI space.
    Useful if the input images already are in MNI space""",
)

outputspace_option = click.option(
    "--space",
    "outputspace",
    type=click.Choice(
        [
            "mr",
            "pet",
        ]
    ),
    default="mr",
    help="Which space to ",
)


@main.command()
@mri_option
@mask_option
@sbpet_output_option
@verbose_option
@device_option
def syn(
    mri_fnames: Sequence[str],
    mask_fnames: Sequence[str],
    out_fnames: Union[Sequence[str], None] = None,
    verbose: bool = True,
    device: str = "cuda:0",
) -> None:
    """Synthesize baseline PET images."""
    if out_fnames is None or len(out_fnames) == 0:
        out_fnames = [
            _create_output_fname(mri_fname, suffix="_sbraw") for mri_fname in mri_fnames
        ]

    synthesize_baselines(
        mri_fnames,
        mask_fnames,
        out_fnames,
        verbose=verbose,
        device=device,
    )


pet_option = click.option(
    "-p",
    "--pet",
    "pet_fnames",
    type=click.Path(exists=True),
    multiple=True,
    required=True,
)

sbpet_option = click.option(
    "-s",
    "--sbpet",
    "sbpet_fnames",
    type=click.Path(exists=True),
    multiple=True,
    required=True,
)

abn_output_option = click.option(
    "-o",
    "--o",
    "out_fnames",
    type=click.Path(),
    multiple=True,
)

no_resample_mask_option = click.option(
    "--no-resample-mask",
    "no_resample_mask",
    is_flag=True,
    default=False,
    help="Print verbose output.",
)


@pet_option
@sbpet_option
@mask_option
@abn_output_option
@verbose_option
@device_option
@main.command()
def abn(
    pet_fnames: Sequence[str],
    sbpet_fnames: Sequence[str],
    mask_fnames: Sequence[str],
    out_fnames: Union[Sequence[str], None] = None,
    verbose: bool = False,
    device: str = "cuda:0",
):
    """Create abnormality maps."""
    if out_fnames is None or len(out_fnames) == 0:
        out_fnames = [
            _create_output_fname(pet_fname, suffix="_abn") for pet_fname in pet_fnames
        ]

    create_abnormality_maps(
        pet_fnames,
        sbpet_fnames,
        mask_fnames,
        out_fnames,
        verbose=verbose,
        device=device,
    )


no_resample_sbpet_option = click.option(
    "--no-resample-sbpet",
    "no_resample_sbpet",
    is_flag=True,
    default=False,
    help="Print verbose output.",
)


@pet_option
@sbpet_option
@mask_option
@abn_output_option
@verbose_option
@device_option
@main.command()
def norm(
    pet_fnames: Sequence[str],
    sbpet_fnames: Sequence[str],
    mask_fnames: Sequence[str],
    out_fnames: Union[Sequence[str], None] = None,
    verbose: bool = False,
    device: str = "cuda:0",
):
    """Normalize sbPET images to PET images."""
    if out_fnames is None or len(out_fnames) == 0:
        out_fnames = [
            _create_output_fname(pet_name, suffix="_sb") for pet_name in pet_fnames
        ]

    normalize_to_pet(
        pet_fnames=pet_fnames,
        sbpet_fnames=sbpet_fnames,
        mask_fnames=mask_fnames,
        out_fnames=out_fnames,
        verbose=verbose,
        device=device,
    )


no_image_option = click.option(
    "--no-image",
    "no_image",
    is_flag=True,
    default=False,
    help="Print verbose output.",
)

mri_option_single = click.option(
    "-i",
    "--mri",
    "mri_fname",
    type=click.Path(exists=True),
    required=True,
)
mask_option_single = click.option(
    "-m",
    "--mask",
    "mask_fname",
    type=click.Path(exists=True),
    required=True,
)

pet_option_single = click.option(
    "-p",
    "--pet",
    "pet_fname",
    type=click.Path(exists=True),
    required=False,
)

sbpet_output_option_single = click.option(
    "-os",
    "--out-sbpet",
    "out_sbpet",
    type=click.Path(),
    required=False,
)

abn_output_option_single = click.option(
    "-oa",
    "--out-abn",
    "out_abn",
    type=click.Path(),
    required=False,
)

img_output_option_single = click.option(
    "-oi",
    "--out-img",
    "out_img",
    type=click.Path(),
    required=False,
)

no_abnormality_option = click.option(
    "--no-abn",
    "no_abnormality",
    is_flag=True,
    default=False,
    help="Print verbose output.",
)

no_normalization_option = click.option(
    "--no-norm",
    "no_normalization",
    is_flag=True,
    default=False,
    help="Print verbose output.",
)

no_image_option = click.option(
    "--no-img",
    "no_image",
    is_flag=True,
    default=False,
    help="Print verbose output.",
)


@mri_option_single
@mask_option_single
@pet_option_single
@sbpet_output_option_single
@abn_output_option_single
@img_output_option_single
@no_registration_option
@no_abnormality_option
@no_normalization_option
@no_image_option
@verbose_option
@device_option
@outputspace_option
@main.command()
def run(
    mri_fname,
    mask_fname,
    pet_fname,
    out_sbpet,
    out_abn,
    out_img,
    no_registration,
    no_abnormality,
    no_normalization,
    no_image,
    verbose,
    device,
    outputspace,
):
    """Run full pipeline."""
    do_registration = not no_registration
    do_abnormality = not no_abnormality
    do_normalization = not no_normalization
    do_image = not no_image

    if out_sbpet is None:
        out_sbpet = _create_output_fname(pet_fname, suffix="_sb")
    if out_abn is None:
        out_abn = _create_output_fname(pet_fname, suffix="_abn")

    run_full(
        mri_fname=mri_fname,
        mask_fname=mask_fname,
        out_sbpet=out_sbpet,
        pet_fname=pet_fname,
        out_abn=out_abn,
        out_img=out_img,
        do_registration=do_registration,
        do_abnormality=do_abnormality,
        do_normalization=do_normalization,
        do_image=do_image,
        verbose=verbose,
        device=device,
        outputspace=outputspace,
    )



"""Niftyreg wrapper for torchio."""
import tempfile

import niftyreg  # type: ignore
import numpy as np

from zerodose.utils import binarize
from zerodose.utils import get_mni_template


def from_mni(
    ref_fname,
    in_affine_fname,
    float_fname,
    out_fname,
):
    """Resample a floating image to a reference image using an affine matrix."""
    with tempfile.NamedTemporaryFile() as temp_inv_affine:
        affine_mat = _read_matrix_nifty(in_affine_fname)
        inv_affine_mat = np.linalg.inv(affine_mat)
        inv_affine = temp_inv_affine.name
        _save_matrix_nifty(inv_affine_mat, inv_affine)

        niftyreg.main(
            [
                "resample",
                "-ref",
                ref_fname,
                "-flo",
                float_fname,
                "-res",
                out_fname,
                "-aff",
                inv_affine,
                "-voff",
            ]
        )


def to_mni(
    in_mri_fname,
    in_mask_fname,
    in_pet_fname,
    out_mri_fname,
    out_mask_fname,
    out_pet_fname,
    out_affine_fname,
):
    """Resample a floating image to a reference image using an affine matrix."""
    ref = get_mni_template()
    niftyreg.main(
        [
            "aladin",
            "-flo",
            in_mri_fname,
            "-ref",
            ref,
            "-res",
            out_mri_fname,
            "-aff",
            out_affine_fname,
            "-speeeeed",
            "-voff",
        ],
    )
    in_affine = out_affine_fname
    niftyreg.main(
        [
            "resample",
            "-ref",
            ref,
            "-flo",
            in_mask_fname,
            "-res",
            out_mask_fname,
            "-aff",
            in_affine,
            "-voff",
        ]
    )

    binarize(out_mask_fname, out_mask_fname)
    if in_pet_fname is not None:
        niftyreg.main(
            [
                "resample",
                "-ref",
                ref,
                "-flo",
                in_pet_fname,
                "-res",
                out_pet_fname,
                "-aff",
                in_affine,
            ]
        )


def _save_matrix_nifty(affine_mat, file_name):
    s = ""
    for i in range(affine_mat.shape[0]):
        for j in range(affine_mat.shape[1]):
            affine_mat[i, j] = affine_mat[i, j].item()
            s += str(affine_mat[i, j]) + " "
        s = s[:-1]
        s += "\n"
    with open(file_name, "w") as f:
        f.write(s)


def _read_matrix_nifty(file_name):
    with open(file_name) as f:
        s = f.read()
    s = s.split("\n")
    s = s[:-1]
    s = [i.split(" ") for i in s]
    s = [[float(j) for j in i] for i in s]
    s = np.array(s)
    return s



"""Inference functions."""
from typing import Tuple
from typing import Union

import torch
import torchio as tio
from torch.utils.data import DataLoader
from torchio.data import GridAggregator
from torchio.data import GridSampler

from zerodose import utils
from zerodose.models import ZeroDose


def _infer_single_subject(
    model: ZeroDose,
    subject: tio.Subject,
    ps: Tuple[int, int, int],
    po: Tuple[int, int, int],
    bs: int,
    std: Union[float, int],
    device: Union[torch.device, str],
) -> torch.Tensor:
    """Infer a single subject."""
    grid_sampler = GridSampler(subject, ps, po)
    patch_loader = DataLoader(
        grid_sampler, batch_size=bs, num_workers=0  # type: ignore
    )
    aggregator = GridAggregator(grid_sampler, overlap_mode="average")
    aggregator_weight = GridAggregator(grid_sampler, overlap_mode="average")
    weight = utils.get_gaussian_weight(ps, std).to(device)
    with torch.no_grad():
        for patches_batch in patch_loader:
            patch_x = patches_batch["mr"][tio.DATA].to(device=device)
            locations = patches_batch[tio.LOCATION]
            patch_y = model(patch_x)
            aggregator.add_batch(patch_y * weight, locations)
            aggregator_weight.add_batch(weight, locations)

    return (
        aggregator.get_output_tensor().detach()
        / aggregator_weight.get_output_tensor().detach()
    )


def synthesize(mr, model, device="cuda:0"):
    """Synthesize baseline PET images from MRI images."""
    if len(mr.shape) == 3:
        mr = mr.unsqueeze(0)

    sub = tio.Subject({"mr": tio.ScalarImage(tensor=mr)})
    stride = 2
    patch_size = (32, 192, 192)
    patch_overlap = (32 - stride, 192 - stride, 192 - stride)
    sd_weight = (5,)
    batch_size = (1,)
    stride = (2,)

    sbpet_tensor = _infer_single_subject(
        model, sub, patch_size, patch_overlap, batch_size, sd_weight, device
    )

    return sbpet_tensor



"""Processing functions for the ZeroDose model."""

import torch
import torchio as tio
from torchio import SpatialTransform
from torchio.data.subject import Subject

from zerodose import processing


_X_MIN_MNI = 0
_X_MAX_MNI = 192
_Y_MIN_MNI = 21
_Y_MAX_MNI = 213
_Z_MIN_MNI = -7
_Z_MAX_MNI = 185


def _crop_mni_to_192(arr: torch.Tensor) -> torch.Tensor:
    """Crop the MNI image to 192x192x192."""
    parr = torch.zeros((1, _X_MAX_MNI - _X_MIN_MNI, _Y_MAX_MNI - _Y_MIN_MNI, 192))
    parr[:, :, :, abs(_Z_MIN_MNI) : 192] = arr[
        :, _X_MIN_MNI:_X_MAX_MNI, _Y_MIN_MNI:_Y_MAX_MNI, :_Z_MAX_MNI
    ]

    return parr


def _crop_192_to_mni(arr: torch.Tensor) -> torch.Tensor:
    """Crop the 192x192x192 image to MNI."""
    parr = torch.zeros((1, 197, 233, 189))
    parr[:, _X_MIN_MNI:_X_MAX_MNI, _Y_MIN_MNI:_Y_MAX_MNI, :_Z_MAX_MNI] = arr[
        :, :, :, abs(_Z_MIN_MNI) : 192
    ]

    return parr


class PadAndCropMNI(SpatialTransform):
    """Pad the MNI image to 192x192x192."""

    def __init__(self, is_inverse=False, **kwargs) -> None:
        """Initialize the transform."""
        self.is_inverse = is_inverse
        super().__init__(**kwargs)

    def apply_transform(self, subject: Subject) -> Subject:
        """Apply the transform to the subject."""
        for image in self.get_images(subject):
            if self.is_inverse:
                _pad_inv(image)
            else:
                _pad(image)

        return subject

    @staticmethod
    def is_invertible() -> bool:
        """Return whether the transform is invertible."""
        return True

    def inverse(self):
        """Returns the inverse transform."""
        return PadAndCropMNI(is_inverse=True)


class Binarize(SpatialTransform):
    """Binarize an image based on a threshhold."""

    def __init__(self, threshold=0.5, **kwargs) -> None:
        """Initialize the transform."""
        self.threshold = threshold
        super().__init__(**kwargs)

    def apply_transform(self, subject: Subject) -> Subject:
        """Apply the transform to the subject."""
        for image in self.get_images(subject):
            data = image.data
            image.set_data(data > self.threshold)
        return subject


def _pad(image: tio.Image) -> None:
    data = processing._crop_mni_to_192(image.data)
    image.set_data(data)


def _pad_inv(image: tio.Image) -> None:
    data = processing._crop_192_to_mni(image.data)
    image.set_data(data)


class ToFloat32(SpatialTransform):
    """Convert the image to float32."""

    def __init__(self, **kwargs) -> None:
        """Initialize the transform."""
        super().__init__(**kwargs)

    def apply_transform(self, subject: Subject) -> Subject:
        """Apply the transform to the subject."""
        for image in self.get_images(subject):
            _to_float(image)

        return subject

    @staticmethod
    def is_invertible() -> bool:
        """Return whether the transform is invertible."""
        return False


def _to_float(image: tio.Image) -> None:
    _data = image.numpy().astype("f")
    data = torch.as_tensor(_data)
    image.set_data(data)



"""Utility functions."""
import math
import os
import re
import shutil
import tempfile
from typing import List
from typing import Literal
from typing import Tuple
from typing import Union
from urllib.request import urlopen

import nibabel as nib
import torch
import torch.nn as nn
from torch.nn import functional

from zerodose.models import ZeroDose
from zerodose.paths import folder_with_parameter_files
from zerodose.paths import folder_with_templates


def binarize(img_path, out_path, threshold=0.5):
    """Binarize an image."""
    img = nib.load(img_path)
    data = img.get_fdata()
    data[data > threshold] = 1
    img = nib.Nifti1Image(data, img.affine, img.header)
    nib.save(img, out_path)


def _get_model_fname() -> str:
    """Returns the path to the parameter file for the given fold."""
    return os.path.join(folder_with_parameter_files, "model_1.pt")


def _get_mni_template_fname():
    return os.path.join(folder_with_templates, "t1.nii")


def _download_file(url: str, filename: str) -> None:
    data = urlopen(url).read()  # noqa
    with open(filename, "wb") as f:
        f.write(data)


def _maybe_download_parameters(
    force_overwrite: bool = False, verbose: bool = True
) -> None:
    """Downloads the parameters for some fold if it is not present yet.

    :param force_overwrite: if True the old parameter file will be\
          deleted (if present) prior to download
    :return:
    """
    if not os.path.isdir(folder_with_parameter_files):
        _maybe_mkdir_p(folder_with_parameter_files)

    out_filename = _get_model_fname()

    if force_overwrite and os.path.isfile(out_filename):
        os.remove(out_filename)

    if not os.path.isfile(out_filename):
        url = "http://sandbox.zenodo.org/record/1165160/files/gen2.pt?download=1"
        if verbose:
            print("Downloading model parameters", url, "...")
        _download_file(url, out_filename)


def _maybe_download_mni_template(
    force_overwrite: bool = False, verbose: bool = True
) -> None:
    if not os.path.isdir(folder_with_templates):
        _maybe_mkdir_p(folder_with_templates)

    out_filename = _get_mni_template_fname()

    if force_overwrite and os.path.isfile(out_filename):
        os.remove(out_filename)

    if not os.path.isfile(out_filename):
        with tempfile.TemporaryDirectory() as tempdirname:
            url = (
                "https://www.bic.mni.mcgill.ca/~vfonov/icbm/"
                "2009/mni_icbm152_nlin_sym_09a_nifti.zip"
            )
            filename = os.path.join(tempdirname, "mni_icbm152_nlin_sym_09a_nifti.zip")
            if verbose:
                print("Downloading MNI template", url, "...")
            _download_file(url, filename)
            if verbose:
                print("Unpacking MNI template", url, "...")
            shutil.unpack_archive(filename, tempdirname)
            t1 = os.path.join(
                tempdirname,
                "mni_icbm152_nlin_sym_09a",
                "mni_icbm152_t1_tal_nlin_sym_09a.nii",
            )
            shutil.copy(t1, out_filename)


def _maybe_mkdir_p(directory: str) -> None:
    """Creates a directory if it does not exist yet."""
    if not os.path.exists(directory):
        os.makedirs(directory)


class GaussianSmoothing(nn.Module):
    """Apply gaussian smoothing on a tensor."""

    def __init__(self, channels, kernel_size, sigma, dim=2):
        """Make a new instance of the GaussianSmoothing class."""
        super().__init__()
        self.kernel_size = kernel_size
        if self.kernel_size % 2 == 0:
            raise ValueError("Kernel size must be odd.")
        kernel_size = [kernel_size] * dim
        sigma = [sigma] * dim

        # The gaussian kernel is the product of the
        # gaussian function of each dimension.
        kernel = 1
        meshgrids = torch.meshgrid(
            [torch.arange(size, dtype=torch.float32) for size in kernel_size],
            indexing="xy",
        )
        for size, std, mgrid in zip(kernel_size, sigma, meshgrids):  # noqa
            mean = (size - 1) / 2
            kernel *= (
                1
                / (std * math.sqrt(2 * math.pi))
                * torch.exp(-(((mgrid - mean) / (2 * std)) ** 2))
            )

        # Make sure sum of values in gaussian kernel equals 1.
        kernel = kernel / torch.sum(kernel)

        # Reshape to depthwise convolutional weight
        kernel = kernel.view(1, 1, *kernel.size())
        kernel = kernel.repeat(channels, *[1] * (kernel.dim() - 1))

        self.register_buffer("weight", kernel)
        self.groups = channels

        if dim == 1:
            self.conv = functional.conv1d
        elif dim == 2:
            self.conv = functional.conv2d
        elif dim == 3:
            self.conv = functional.conv3d
        else:
            raise RuntimeError(
                f"Only 1, 2 and 3 dimensions are supported. Received {dim}."
            )

    def forward(self, input):
        """Forward pass of the gaussian smoothing."""
        return self.conv(
            input.unsqueeze(0),
            weight=self.weight,
            groups=self.groups,
            padding=(self.kernel_size - 1) // 2,
        ).squeeze(0)


def get_gaussian_weight(
    ps: Union[List[int], Tuple[int, int, int]], std: Union[float, int]
) -> torch.Tensor:
    """Returns a gaussian weight for the given patch size and standard deviation."""
    n_slices = ps[0]
    gaussian_vec = (
        1
        / torch.sqrt(torch.tensor(std) * torch.pi)
        * torch.exp(
            -0.5 * (torch.square(torch.arange(n_slices) + 0.5 - n_slices / 2.0) / std)
        )
    )
    gaussian_vec = gaussian_vec.unsqueeze(1).unsqueeze(1)
    weight = torch.ones(ps)
    weight *= gaussian_vec
    return weight.unsqueeze(0).unsqueeze(0)


def save_nifty(data: torch.Tensor, filename_out: str, affine_ref: str) -> None:
    """Saves a nifty file."""
    save_directory = os.path.dirname(filename_out)
    if not os.path.exists(save_directory):
        os.makedirs(save_directory)

    func = nib.load(affine_ref).affine
    data = data.squeeze().cpu().detach().numpy()
    ni_img = nib.Nifti1Image(data, func)
    nib.save(ni_img, filename_out)


def load_nifty(fname: str) -> torch.Tensor:
    """Loads a nifty file and returns it as a torch tensor."""
    return torch.from_numpy(nib.load(fname).get_fdata()).float()


def get_mni_template() -> str:
    """Returns the path to the t1 MNI template. If missing, downloads the template."""
    if os.environ.get("GITHUB_ACTIONS"):
        raise Exception(
            """MNI templates may not be downloaded by
            pytests that do not have a 'slow' marker"""
        )
    else:
        _maybe_download_mni_template()
        return _get_mni_template_fname()


def get_model(
    model_type: Literal["standard", "dummy", "determine"] = "determine"
) -> ZeroDose:
    """Returns the ZeroDose model and loads the parameters."""
    if model_type == "determine":
        if os.environ.get("ZERODOSE_USE_DUMMY_MODEL"):
            model_type = "dummy"
        else:
            model_type = "standard"

    if model_type == "dummy":
        return ZeroDose(model_type=model_type)
    elif model_type == "standard":
        model = ZeroDose(model_type=model_type)
        if not os.environ.get("GITHUB_ACTIONS"):
            _maybe_download_parameters()
            weights_path = _get_model_fname()
            model.generator.load_state_dict(torch.load(weights_path))
        return model
    else:
        raise ValueError(f"Unknown model type '{model_type!r}'.")


def _create_output_fname(mri_fname, suffix="_sb", file_type=".nii.gz"):
    """Create output filename from input filename."""
    out_fname = mri_fname
    if out_fname.endswith(".nii.gz"):
        out_fname = re.sub(".nii.gz$", "", out_fname)
    if out_fname.endswith(".nii"):
        out_fname = re.sub(".nii$", "", out_fname)
    out_fname += suffix + file_type
    return out_fname
